use std::{fs::File, io::{self, BufWriter, Write}};

mod obj_parse;
mod optimize;


#[derive(Debug,PartialEq,Eq,Clone)]
struct QVertex {
    x: i16, y: i16, z: i16,
    u: u16, v: u16,
    nx: i8, ny: i8, nz: i8,
    shade: u8,
}

#[derive(Debug,Clone)]
struct F32Vertex {
    x: f32, y: f32, z: f32,
    u: f32, v: f32,
    nx: f32, ny: f32, nz: f32,
    shade: f32,
}



fn quantize_vertices(vt: &[F32Vertex],offset: (f32,f32,f32), scale: (f32,f32,f32), tex_scale: (f32,f32)) -> Vec<QVertex> {
    // Note here that "f32 as i16" etc inherently clamp in Rust
    vt.iter().map(|v| {
        QVertex {
            x: ((v.x + offset.0) * scale.0).round() as i16,
            y: ((v.y + offset.1) * scale.1).round() as i16,
            z: ((v.z + offset.2) * scale.2).round() as i16,
            u: (v.u * tex_scale.0).round() as u16,
            v: (v.v * tex_scale.1).round() as u16,
            nx: (v.nx * 127.0).round() as i8,
            ny: (v.ny * 127.0).round() as i8,
            nz: (v.nz * 127.0).round() as i8,
            shade: (v.shade * 255.0).round() as u8,
        }
    }).collect()
}

fn write_old_format(file: &mut dyn Write,tris: &[[u32;3]], verts: &[QVertex]) -> io::Result<()> {
    let mut buf = BufWriter::new(file);
    buf.write(&(verts.len() as u16).to_le_bytes())?;
    buf.write(&(tris.len() as u16).to_le_bytes())?;
    for v in verts {
        buf.write(&(v.x).to_le_bytes())?;
        buf.write(&(v.y).to_le_bytes())?;
        buf.write(&(v.z).to_le_bytes())?;
        buf.write(&(v.nx).to_le_bytes())?;
        buf.write(&(v.ny).to_le_bytes())?;
        buf.write(&(v.nz).to_le_bytes())?;
        buf.write(&(v.shade).to_le_bytes())?;
    }
    for t in tris {
        buf.write(&(t[0] as u16).to_le_bytes())?;
        buf.write(&(t[1] as u16).to_le_bytes())?;
        buf.write(&(t[2] as u16).to_le_bytes())?;
    }
    Ok(())
}

fn main() {
    let model = obj_parse::parse_file(&mut File::open("../teapot3.obj").unwrap());
    let q_verts = quantize_vertices(&model.1, (0.,0.,0.), (24576.0,24576.0,24576.0), (0xffff as f32,0xffff as f32));

    println!("Faces: {}, Verts: {}",model.0.len(),model.1.len());

    write_old_format(&mut File::create("../teapot3_r.dat").unwrap(), &model.0, &q_verts).unwrap();

    let cb = optimize::geo_encode(&model.0,&q_verts);
    File::create("../teapot3_geocmd.dat").unwrap().write(cb.as_slice()).unwrap();
}
